/*
from pt3md_fmt.txt:
---------------------------------------- 
 :

     -  ,  -
    -  , ,  ...
      .    -
   1 .

)  A..C

 :
 [PP] NN XV II OO DC EE [eoe]
 
  :
 PP     -  (  )
 NN     - 
 XV     - \   
 II     - 
 OO     - 
 DC     -   
 EE     -  .
 [eoe]  -   (end of event) ( )
*/

/*
 PP - :
        0x64 -  
        0x65 -   ----//---
        0x70..0x7F -   
                
        0x67 -   ,  -  + 1,
                    
        0x68..0x6F - ,      
                 1..8
               
     7   .
*/
typedef u8 pmd_chan_event_PP;

/*
 NN - bit 0..6: 0    -   (...)
                1-96 -    C-1...B-8
                97   - note off (S--)
                98   - note cut (R--)
                
          7:    0    -  
                1    -  XV   .
*/

enum pmd_chan_event_NN_note{
	noteoff = 97, notecut
};

typedef struct {
	u8 note:7;
	u8 notlast:1;
} pmd_chan_event_NN;

/*
 XV - bit 0..3:    0..15
          4:      (1 - )
          
           0..4     " " 
               (..  ), .:
          
           new_volume = old_volume ^ (XV & 0x1F);
          
          5:       II ( )
          6:       OO ( )
          7:       DC (  ), 
                EE ( )   -   .
*/

typedef struct {
	u8 vol:4;
	u8 env:1;
	u8 ins:1;
	u8 orn:1;
	u8 cmd:1;
} pmd_chan_event_XV;

/*
 II - bit 0..7:  
*/
typedef u8 pmd_chan_event_II;

/*
 OO - bit 0..7:  
*/
typedef u8 pmd_chan_event_OO;

/*
 DC - bit 0..3:  
          4..7:  
          
         ,      
*/
typedef struct {
	u8 cmd:4;
	u8 subparm:4;
} pmd_chan_event_DC;

/*
 EE - bit 0..7:  
*/
typedef u8 pmd_chan_event_EE;

/*
 )  \

 :
 [XX] XT PP OO NN DC EE [eoe]
 
  :
 XX     -  (  )
 XT     -  \   
 PP     -     
 OO     -     
 NN     -  
 DC     -   
 EE     -  .
 [eoe]  -   (end of event) ( )
*/

/*
 XX -   3..7 = 0:
    0    -   - -  
    1..7 - ,        1..7
*/

typedef u8 pmd_env_event_XT;

/*              
 XT - bit 0..2:  bit 3 = 0 - 
                 bit 3 = 1 -     8..F
                ( 0..3     9, 4..7 - F)
          3:    1 -    (), 0 -  
          4..5:   PP OO:
                0: PP  OO ,   
                1: PP -   ,  = 0, OO 
                2: PP -   , OO - 
                3: PP - , OO -  ( PP bit 7 = 1, . )
          6:       NN ( )
          7:       DC (  ), 
                EE ( )   -   .
*/
enum pmd_env_event_XT_fmt{
	none, period_lsb, period, note
};

typedef union {
	struct {
		u8 type:3;
		u8 restart:1;

		u8 fmt:2;

		u8 noise:1;
		u8 cmd:1;
	};
	u8 val;
} pmd_env_event_XT;
/*
 PP -  XT bit 4..5 = 3:
      bit 0..6: 0    -   (...)
                1-96 -    C-1...B-8
          7:    0    -   
                1    -  OO   .
      
       XT bit 4..5 = 1,2:
      bit 0..7:    .
*/
typedef union {
	struct {
		u8 note:7;
		u8 orn:1;
	};
	u8 period_lsb;
} pmd_env_event_PP;

/*
 OO -  (XT bit 4..5 = 3) && (PP bit 7 = 1):
      bit 0..7:   (0 - )
      
       XT bit 4..5 = 2:
      bit 0..7:    .
*/
typedef u8 pmd_env_event_OO;

/*
 NN - bit 0..4:   (),    
                  NN
*/
typedef u8 pmd_env_event_NN;

/*
 DC - bit 0..3:  
          4..7:  
          
         ,      
*/
typedef struct {
	u8 cmd:4;
	u8 subparm:4;
} pmd_env_event_DC;

/*
 EE - bit 0..7:  
*/
typedef u8 pmd_env_event_EE;
