#include "stdafx.h"
#include "plrint.h"

// call at start
void cmd_set(int ch, int num, int subparm, int parm) {
	signed int rawslide;

	switch(num) {
	case 0:	// no cmd
		break;

	case 1:
	case 2: // period slide
		if (!pmdmdl->flags.tremor) chan[ch].cmd.tremor.active = false;

		rawslide = (num == 2 ? -parm : parm);
		if (subparm) {
			chan[ch].cmd.slide.active = true;
			chan[ch].cmd.slide.porta  = false;
			chan[ch].cmd.slide.add    = rawslide;
			chan[ch].cmd.slide.delay  = chan[ch].cmd.slide.count = subparm;
		} else {
			chan[ch].cmd.slide.active = false;
			chan[ch].cmd.slide.acc   += rawslide;	// emulate pt3.7 behavior
		}
		break;

	case 3:	// portamento
		signed int start, end;
		if (!pmdmdl->flags.tremor) chan[ch].cmd.tremor.active = false;

		if (subparm) {
			chan[ch].cmd.slide.active = chan[ch].cmd.slide.porta = true;
			chan[ch].flushacc = false;

			if (ch < 3) {
				start = notetab[chan[ch].oldnote];
				end   = notetab[chan[ch].basenote];
				chan[ch].cmd.slide.portanote = chan[ch].basenote; chan[ch].basenote = chan[ch].oldnote;
			} else {
				// env chan code
				start = (chan[ch].oldnote  != -1 ? notetab[chan[ch].oldnote]  : env_chan.env.oldbase);
				end   = (chan[ch].basenote != -1 ? notetab[chan[ch].basenote] : env_chan.env.base);
				env_chan.env.portabase = env_chan.env.base; env_chan.env.base = env_chan.env.oldbase;
				chan[ch].basenote = chan[ch].oldnote = -1; // hahahaha! just make code runs faster
			}

			chan[ch].cmd.slide.limit = end - start;
			chan[ch].cmd.slide.add   = ((chan[ch].cmd.slide.limit - chan[ch].cmd.slide.acc) < 0 ? -parm : parm);
			chan[ch].cmd.slide.delay = chan[ch].cmd.slide.count = subparm;
		} else {
			chan[ch].cmd.slide.active = false;
		}
		break;

	case 4:	// instrument offset
		ins_set(ch, chan[ch].ins.num, parm);
		break;

	case 5:	// ornament offset
		orn_set(ch, chan[ch].orn.num, parm);
		break;

	case 6:	// tremor \ note cut \ note delay
		if (!pmdmdl->flags.tremor) chan[ch].cmd.slide.active = false;

		// note cut is not implemented
		if (parm) {
			chan[ch].cmd.tremor.active  = chan[ch].cmd.tremor.dir = true;
			chan[ch].cmd.tremor.ontime  = chan[ch].cmd.tremor.count = (parm >> 4);
			chan[ch].cmd.tremor.offtime = (parm & 0xF);
			chan[ch].cmd.tremor.count++;
		} else chan[ch].cmd.tremor.active = false;
		break;

	case 7: // channel tone\noise\env mask
		// new pmd 0.3+ interpretation
		chan[ch].globmixer.tone  = ((~subparm & 0x1));
		chan[ch].globmixer.noise = ((~subparm & 0x8) >> 3);
		chan[ch].globmixer.env   = ( (subparm & 0x2) >> 1);
		break;

	case 8:	// volume slide
		if (subparm) {
			chan[ch].cmd.volslide.active = true;
			//    x\y (   parm   )
			chan[ch].cmd.volslide.delay  = chan[ch].cmd.volslide.count = subparm;

			unsigned int x = (parm >> 4), y = (parm & 0xF);
			if ((x == 0) && (y > 0)) chan[ch].cmd.volslide.add = -(int)y; else
			if ((y == 0) && (x > 0)) chan[ch].cmd.volslide.add = x; else chan[ch].cmd.volslide.active = false;
		} else chan[ch].cmd.volslide.active = false;
		break;

	case 0xF:	// custom commands
		switch(subparm) {
		case 0:	// set speed
			if (parm) {
				player.speed = player.spdpos = parm;
				player.autospd.active = false;
			} else {
				player.speed = player.spdpos = player.autospd.spd;
				player.autospd.active = true;
			}
			break;

		default:
			break;
		}
		break;

	default:
		chan[ch].cmd.num = chan[ch].cmd.parm = chan[ch].cmd.subparm =  -1;
		break;
	}
}

// call every tick
void cmd_process(int ch) {

	// process slides
	if (chan[ch].cmd.slide.active) {
		if (!(--chan[ch].cmd.slide.count)) {
			// update
			chan[ch].cmd.slide.count = chan[ch].cmd.slide.delay;

			chan[ch].cmd.slide.acc += chan[ch].cmd.slide.add;

			if  ((chan[ch].cmd.slide.porta) &&
			   (((chan[ch].cmd.slide.add >= 0) && (chan[ch].cmd.slide.acc >= chan[ch].cmd.slide.limit)) ||
			    ((chan[ch].cmd.slide.add <  0) && (chan[ch].cmd.slide.acc <= chan[ch].cmd.slide.limit)))) {
				// disable portamento
				if (ch < 3) chan[ch].basenote = chan[ch].cmd.slide.portanote; else
					env_chan.env.base = env_chan.env.portabase;
				chan[ch].cmd.slide.active = false;
				chan[ch].cmd.slide.acc    = 0;
			}
		}
	}

	// process tremor
	if (chan[ch].cmd.tremor.active) {
		if (!(--chan[ch].cmd.tremor.count)) {
			// update
			chan[ch].cmd.tremor.dir = !chan[ch].cmd.tremor.dir;
			chan[ch].cmd.tremor.count = (chan[ch].cmd.tremor.dir ? chan[ch].cmd.tremor.ontime : chan[ch].cmd.tremor.offtime);

			chan[ch].off = !chan[ch].cmd.tremor.dir;
		}
	}

	// process volume slides
	if (chan[ch].cmd.volslide.active) {
		if (!(--chan[ch].cmd.volslide.count)) {
			// update
			chan[ch].cmd.volslide.count  = chan[ch].cmd.volslide.delay;
			chan[ch].volume_ch = clamp(chan[ch].volume_ch + chan[ch].cmd.volslide.add, 0, 15);
		}
	}
}
