#include "stdafx.h"
#include "plrint.h"

// call at start
void ins_set(int ch, int num, int pos) {
	// check for empty or absent instrument
	if (*((u16*)&pmdstream[player.instab[num]]) == 0) {chan[ch].ins.num = 0; return;}

	pt3md_ins_hdr *tmphdr = (pt3md_ins_hdr*)&pmdstream[player.instab[num]];
	if (tmphdr->linesize == 0)  {chan[ch].ins.num = 0; return;}

	// set new instrument data
	chan[ch].ins.num = num;
	chan[ch].ins.hdr = (pt3md_ins_hdr*)&pmdstream[player.instab[num]];
	chan[ch].ins.ftr = (pt3md_ins_ftr*)&pmdstream[player.instab[num] + (chan[ch].ins.hdr->len * chan[ch].ins.hdr->linesize) + sizeof(pt3md_ins_hdr)];

	// stop slide cmds
	chan[ch].cmd.slide.active = false;
	// stop tremor
	chan[ch].cmd.tremor.active = false;
	// stop volume slide
	chan[ch].cmd.volslide.active = false;

	// flush accumulators
	chan[ch].ins.volbase = chan[ch].ins.t_acc = chan[ch].ins.n_acc = chan[ch].ins.e_acc = 0;

	// parse header
	chan[ch].ins.sustain = (bool)chan[ch].ins.hdr->sustain;
	chan[ch].ins.rep     = &pmdstream[player.instab[num] + sizeof(pt3md_ins_hdr) + (chan[ch].ins.hdr->loop_start * chan[ch].ins.hdr->linesize)];
	
	if (pos > (chan[ch].ins.hdr->len - 1)) pos = chan[ch].ins.hdr->loop_start;
	chan[ch].ins.dat     = &pmdstream[player.instab[num] + sizeof(pt3md_ins_hdr) + (pos * chan[ch].ins.hdr->linesize)];
	chan[ch].ins.pos	 = pos;

	if (chan[ch].ins.sustain) {
		chan[ch].ins.s_start = &pmdstream[player.instab[num] + sizeof(pt3md_ins_hdr) + (chan[ch].ins.ftr->loop1_start * chan[ch].ins.hdr->linesize)];
		chan[ch].ins.s_end   = &pmdstream[player.instab[num] + sizeof(pt3md_ins_hdr) + (chan[ch].ins.ftr->loop1_end * chan[ch].ins.hdr->linesize)];
	}
}

// call every tick
void ins_process(int ch) {
	if (ch > 2) return; // invalid for envelope channel
	if (chan[ch].ins.num == 0) {
		chan[ch].mixer.val = 0;
		chan[ch].volume = 0;
		return;
	}

	pt3md_ins_data *rawdata = (pt3md_ins_data*)chan[ch].ins.dat;

	// masks
	chan[ch].mixer.tone  = rawdata->t_mask;
	chan[ch].mixer.noise = rawdata->n_mask;
	chan[ch].mixer.env   = rawdata->e_mask & chan[ch].env;

	signed int rawtone = ((signed int)((rawdata->t_ofsh << 8) | (rawdata->t_ofsl)) ^ 0x800) - 0x800; // sign extension
	
	//    pt3   ,   
	chan[ch].tone.ins = 0;
	if (rawdata->t_acc)  chan[ch].ins.t_acc += rawtone; else chan[ch].tone.ins = rawtone;
	chan[ch].tone.ins += chan[ch].ins.t_acc; // fix:    

	// noise
	signed int rawnoise = ((signed int)((rawdata->n_ofs) ^ 0x10)) - 0x10; // sign extension
	
	// pt3 noise -      
	if (!rawdata->n_mask) {
		if (rawdata->n_acc) {
			chan[ch].ins.n_acc += rawnoise;
			env_chan.noise.acc  = chan[ch].ins.n_acc;
		} else {
			env_chan.noise.acc  = rawnoise;
		    env_chan.noise.acc += chan[ch].ins.n_acc; // fix:    
		}
	}

	// volume
	switch (rawdata->v_slide) {
	case 2:
		chan[ch].ins.volbase--;
		break;
	case 3:
		chan[ch].ins.volbase++;
		break;
	default:
		break;
	}

	chan[ch].volume = voltab[((chan[ch].volume_ch & 0xF) << 4) | clamp(rawdata->volume + chan[ch].ins.volbase, 0, 15)];

	if (chan[ch].ins.hdr->linesize >= 5) {
		// envelope
		signed int rawenv = rawdata->e_ofs;
		if (rawdata->e_acc) chan[ch].ins.e_acc += rawenv; else env_chan.env.acc += rawenv;
		env_chan.env.acc += chan[ch].ins.e_acc;

	}

	// adjust pointer
	chan[ch].ins.dat += (sizeof(u8) * chan[ch].ins.hdr->linesize); chan[ch].ins.pos++;

	if (chan[ch].ins.sustain) {
		if (chan[ch].ins.pos > chan[ch].ins.ftr->loop1_end) {
			chan[ch].ins.pos = chan[ch].ins.ftr->loop1_start;
			chan[ch].ins.dat = chan[ch].ins.s_start;
		}
	} else {
		if (chan[ch].ins.pos >= chan[ch].ins.hdr->len) {
			chan[ch].ins.pos = chan[ch].ins.hdr->loop_start;
			chan[ch].ins.dat = chan[ch].ins.rep;
		}
	}
}

// call at start
void orn_set(int ch, int num, int pos) {
	// check for empty ornament
	if (*((u16*)&pmdstream[player.orntab[num]]) == 0) {chan[ch].orn.num = 0; return;}

	chan[ch].orn.num = num;
	chan[ch].orn.hdr = (pt3md_orn*)&pmdstream[player.orntab[num]];

	// parse header
	chan[ch].orn.rep = (i8*)&pmdstream[player.orntab[num] + sizeof(pt3md_orn) + (chan[ch].orn.hdr->loop_start * sizeof(i8))];
	
	if (pos > (chan[ch].orn.hdr->len - 1)) pos = chan[ch].orn.hdr->loop_start;
	chan[ch].orn.dat = (i8*)&pmdstream[player.orntab[num] + sizeof(pt3md_orn) + (pos * sizeof(i8))];
	chan[ch].orn.pos = pos;

}

// call every tick
void orn_process(int ch) {

	//chan[ch].note = clamp(((signed int)chan[ch].basenote + *chan[ch].orn.dat++), 0, 95); chan[ch].orn.pos++;
	chan[ch].note = chan[ch].basenote + *chan[ch].orn.dat++; if (chan[ch].note < 0) chan[ch].note = 0; chan[ch].orn.pos++;
	if (chan[ch].orn.pos >= chan[ch].orn.hdr->len) {
		chan[ch].orn.pos = chan[ch].orn.hdr->loop_start;
		chan[ch].orn.dat = chan[ch].orn.rep;
	}
}
