#pragma once
#include "player.h"

// player structs
#define sgn(a)			((a) <  0  ?  -1 : ((a) >  0  ?  1  :  0)) 
#define max(a, b)		((a) > (b) ? (a) : (b))
#define min(a, b)		((a) < (b) ? (a) : (b))
#define clamp(a, l, h)	((a) > (h) ? (h) : ((a) < (l) ? (l) : (a)))

#define u8  unsigned char
#define i8  signed char
#define u16 unsigned short
#define i16 signed short

#include "pmd_events.h"
#include "pt3md.h"

// ayregs struct
struct ayregs_t {
	unsigned int val;
	bool dirty;
};

// mixer data
union mixer_t {
	struct{
		unsigned int tone:1;
		unsigned int noise:1;
		unsigned int env:1;
	};
	unsigned int val;
};

// channel structs
struct chan_t {
	unsigned int note, basenote, oldnote;
	
	struct {
		signed   int base;		// period base (note + ornament)

		signed   int ins;		// after instrument
		signed   int final;		// final period	(final = base + slide + ins)

	} tone;

	unsigned int	volume_ch;	// channel volume
	unsigned int    volume;		// final volume

	unsigned int	env;		// envelope mask

	bool            flushacc;	// true - flush accumulator (for proper cmd3 handling)
	bool            cut;		// true - note is cutted
	bool			off;		// true - note is off (goto release)

	mixer_t			mixer;		// mixer masks
	mixer_t			globmixer;	// global mixer masks (by cmd 7)

	struct {
		unsigned int num;
		unsigned int subparm;
		unsigned int parm;

		struct {
			bool active;
			bool dir;			// true - on

			unsigned int count;

			unsigned int ontime;
			unsigned int offtime;
		} tremor;

		struct {
			bool active;
			bool porta;			// true if cmd3, false if cmd1/2
			
			unsigned int portanote;

			unsigned int count;
			unsigned int delay;

			signed   int acc;
			signed   int add;
			signed   int limit; // for cmd 3
			bool         flush; // accumulator flush flag
		} slide;

		struct {
			bool active;
			
			unsigned int count;
			unsigned int delay;

			signed   int acc;
			signed   int add;
		} volslide;

	} cmd;	// for repeating commands

	struct {
		size_t num;
		pt3md_ins_hdr *hdr;
		pt3md_ins_ftr *ftr;

		unsigned int pos;
		bool sustain;

		// DYNAMIC CAST FTW \o/
		u8 *dat;
		u8 *s_start;
		u8 *s_end;
		u8 *rep;

		// data

		// tone\noise\env period accumulators
		signed   int t_acc;
		signed   int n_acc;
		signed   int e_acc;

		signed   int volbase;	// volume base

	} ins;

	struct {
		size_t num;
		pt3md_orn *hdr;
		unsigned int pos;
		i8 *rep;
		i8 *dat;
	} orn;
};

struct env_chan_t {
	struct {
		unsigned int	type;		// env type
		bool			restart;	// true - set new type

		signed   int    base;		// base period
		signed   int    oldbase;
		signed   int    portabase; 

		signed   int    acc;		// instrument accumulator
		signed   int    final;		// final
	} env;

	struct {
		unsigned int	base;		// by env chan
		unsigned int    acc;		// after instruments
		unsigned int    final;
	} noise;
};

// player internal struct
struct player_t {
	struct {
		pt3md_pattern *dat;
		unsigned int num;
		unsigned int pos;
		unsigned int ordpos;
	} pattern;

	struct {
		u8 *pchan;
		
		bool isskip;
		unsigned int skip;
		unsigned int skippos;
	} chan[4];

	struct {
		bool en, active;
		unsigned int pos, len, spd;
		u8 *dat;
		u8 *rep;
	} autospd;

	u16 *instab, *orntab;
	u8  *ordtab, *pattab;

	// speeeeed
	unsigned int speed;
	unsigned int spdpos;

};

// shared resources
#define NOTES	12
#define	OCTAVES	(8+2)		// for envelope as note
extern u8 *pmdstream;
extern unsigned int notetab[NOTES * OCTAVES];
extern unsigned int voltab[16*16];
extern ayregs_t ayregs[256];
extern chan_t chan[4];
extern env_chan_t env_chan;
extern player_t player;

extern pt3md_hdr *pmdhdr;
extern pt3md_mdl_hdr *pmdmdl;

// shared procedures
void write_ay(int r, int d);
void push_ay();

void ins_set(int ch, int num, int pos);
void ins_process(int ch);

void orn_set(int ch, int num, int pos);
void orn_process(int ch);

void cmd_set(int ch, int num, int subparm, int parm);
void cmd_process(int ch);

bool pat_set(int num);
void pat_process();