/*
from pt3md_fmt.txt:
---------------------------------------- 
 :

     -  ,  -
    -  , ,  ...
      .    -
   1 .

)  A..C

 :
 NN XV II OO DC EE [eoe]
 
  :
 NN     - 
 XV     - \   
 II     - 
 OO     - 
 DC     -   
 EE     -  .
 [eoe]  -   (end of event) ( )
*/

/*
 NN - bit 0..6: 0    -   (...)
                1-96 -    C-1...B-8
                97   - note off (S--)
                98   - note cut (R--)
                127  -         XV
                        (. )
          7:    0    -  
                1    -  XV   .
*/

enum pmd_chan_event_NN_note{
	noteoff = 97, notecut, skip = 127 
};

typedef union {
	struct {
		u8 note:7;
		u8 notlast:1;
	};
	u8 val;
} pmd_chan_event_NN;

/*
 XV - bit 0..3:    0..15
          4:    0 -    0..3 XV
                1 -    ()
          5:       II ( )
          6:       OO ( )
          7:       DC (  ), 
                EE ( )   -   .
      
       (NN & 127) == 127,   XV   ,  
          .
*/

typedef struct {
	u8 vol:4;
	u8 env:1;
	u8 ins:1;
	u8 orn:1;
	u8 cmd:1;
} pmd_chan_event_XV;

/*
 II - bit 0..7:  
*/
typedef u8 pmd_chan_event_II;

/*
 OO - bit 0..7:  
*/
typedef u8 pmd_chan_event_OO;

/*
 DC - bit 0..3:  
          4..7:  
          
         ,      
*/
typedef struct {
	u8 cmd:4;
	u8 subparm:4;
} pmd_chan_event_DC;

/*
 EE - bit 0..7:  
*/
typedef u8 pmd_chan_event_EE;

/*
 )  \

 :
 XT PP OO NN DC EE [eoe]
 
  :
 XT     -  \   
 PP     -     
 OO     -     
 NN     -  
 DC     -   
 EE     -  .
 [eoe]  -   (end of event) ( )
*/

/*              
 XT - bit 0..2:     8..F
                ( 0..3     9, 4..7 - F)
          3:    1 -    (), 0 -  
          4..5:   PP OO:
                0: PP  OO ,   
                1: PP -   ,  = 0, OO 
                2: PP -   , OO - 
                3: PP - , OO -  ( PP bit 7 = 1, . )
          6:       NN ( )
          7:       DC (  ), 
                EE ( )   -   .
                
           (!) - XT = 0x0F,     PP 
           ,      , 
           .
*/
enum pmd_env_event_XT_fmt{
	none, period_lsb, period, note
};

typedef union {
	struct {
		u8 type:3;
		u8 restart:1;

		u8 fmt:2;

		u8 noise:1;
		u8 cmd:1;
	};
	u8 val;
} pmd_env_event_XT;
/*
 PP -  XT bit 4..5 = 3:
      bit 0..6: 0    -   (...)
                1-96 -    C-1...B-8
          7:    0    -   
                1    -  OO   .
      
       XT bit 4..5 = 1,2:
      bit 0..7:    .
      
       XT = 0xF:
       ,      .
*/
typedef union {
	struct {
		u8 note:7;
		u8 orn:1;
	};
	u8 period_lsb;
} pmd_env_event_PP;

/*
 OO -  (XT bit 4..5 = 3) && (PP bit 7 = 1):
      bit 0..7:   (0 - )

       XT bit 4..5 = 2:
      bit 0..7:    .
*/
typedef u8 pmd_env_event_OO;

/*
 NN - bit 0..4:   (),    
                  NN
*/
typedef u8 pmd_env_event_NN;

/*
 DC - bit 0..3:  
          4..7:  
          
         ,      
*/
typedef struct {
	u8 cmd:4;
	u8 subparm:4;
} pmd_env_event_DC;

/*
 EE - bit 0..7:  
*/
typedef u8 pmd_env_event_EE;
