#include "pmdconv.h"
#include "optimize.h"

size_t optpos, asdpos;

//     
// find most fitting pattern
std::vector<int> asd_findpattern(std::vector<int> inspd) {

	struct en {
		std::vector<int> d;
		size_t pen;				// penalty count
	};

	std::vector<en> pat;

	// slow slow slow and a bit buggy algo

	for (size_t a = (inspd.size() / 2); a > 0; a--) {
		en v;
		v.d.resize(a);

		std::vector<int>::iterator head = inspd.begin(), tail = inspd.begin() + a;
		std::copy(head, tail, v.d.begin());

		v.pen = 0;
		
		if (verbose) {
			// display pattern
			wprintf(L"--------\n");
			wprintf(L"patlen = %d\n", v.d.size());
			wprintf(L"data: ");
			for (size_t i = 0; i < v.d.size(); i++) wprintf(L"%d,", v.d.at(i));
			wprintf(L"\n");
		}
		
		// do penalty count
		for (size_t i = 0; i < inspd.size(); i++) if (v.d.at(i % v.d.size()) != inspd.at(i)) v.pen++;
		if (verbose) wprintf(L"%d penalties\n", v.pen);

		pat.push_back(v);
	}

	// find most suitable pattern

	std::vector<en>::iterator kewl = pat.end()-1;
	for (std::vector<en>::iterator it = pat.end()-1; it > pat.begin(); it--) {
		if ((it->pen < kewl->pen) && (it->d.size() <= (max((inspd.size() / 16), 1)*kewl->d.size()))) kewl = it;
	}

	if (verbose) {
		// display most suitable pattern
		wprintf(L"-----most suitable pattern----------\n");
		wprintf(L"patlen = %d\n", kewl->d.size());
		wprintf(L"data: ");
		for (size_t i = 0; i < kewl->d.size(); i++) wprintf(L"%d,", kewl->d.at(i));
		wprintf(L"\n");
	}

	if ((kewl->d.size() * 3) < inspd.size()) return kewl->d; else {
		// if pattern is useless - return empty pattern
		if (verbose) wprintf(L"useless - return empty pattern");
		std::vector<int> r(0);
		return r;
	}
}

void spd_push(int spd, pt3md_pattern_data *pat) {
	if (pat->env_chan.cmd == -1) {
		// push in env chan
		pat->env_chan.cmd     = 0xF;
		pat->env_chan.subparm = 0;
		pat->env_chan.parm    = spd;
	} else for (size_t i = 0; i < 3; i++) {
		if (pat->chan[i].cmd == -1) {
			pat->chan[i].cmd     = 0xF;
			pat->chan[i].subparm = 0;
			pat->chan[i].parm    = spd;
		}
	}
}

//  
bool asd_optimize() {
	// data pos accumulator
	asdpos = 0;
	
	wprintf(L"optimize autospeed: ");
	
	// clear storage
	optasd.clear();
	
	for (size_t k = 0; k < pmdmdl->pat_num; k++) {
		if (!pmdpat[k].avail) continue;
		if (!pmdpat[k].autospd.p.size()) continue; // no autospeed
		
		// search for matching pattern
		size_t j = 0;
		while ((j < optasd.size()) && (optasd[j].hash != pmdpat[k].autospd.hash)) j++;

		if (j >= optasd.size()) {
			//   
			optasd.push_back(pmdpat[k].autospd);
			optasd[optasd.size()-1].optidx = asdpos;
			pmdpat[k].autospd.optidx = optasd.size()-1;
			asdpos += (sizeof(u8) * pmdpat[k].autospd.p.size());
		} else pmdpat[k].autospd.optidx = j;	//    
		wprintf(L".");
	}
	wprintf(L"\n");

	return false;
}

//     
bool pat_optimize() {
	// data pos accumulator
	optpos = 0;
	
	wprintf(L"optimize patterns: ");
	
	// clear storage
	optpat.clear();
	
	for (size_t k = 0; k < pmdmdl->pat_num; k++) {
		if (!pmdpat[k].avail) continue;
		for (size_t i = 0; i < 4; i++) {
			// search for matching pattern
			size_t j = 0;
			while ((j < optpat.size()) && (optpat[j].hash != pmdpat[k].data[i].hash)) j++;

			if (j >= optpat.size()) {
				//   
				optpat.push_back(pmdpat[k].data[i]);
				optpat[optpat.size()-1].optidx = optpos;
				pmdpat[k].data[i].optidx = optpat.size()-1;
				optpos += (sizeof(u8) * pmdpat[k].data[i].p.size());
			} else pmdpat[k].data[i].optidx = j;	//    
		}
		wprintf(L".");
	}
	wprintf(L"\n");

	//   ,    :)
	return false;
}
