#include "stdafx.h"
#include "string.h"
#include "stdio.h"
#include "math.h"
#include "ctype.h"

#include "vector"

#include "pmdconv.h"

// pt3.5+ tone tables
const u16 pt3_tonetab[4][12] = {
	{0x0C22, 0x0B73, 0x0ACF, 0x0A33, 0x09A1, 0x0917, 0x0894, 0x0819, 0x07A4, 0x0737, 0x06CF, 0x066D},	// table 0
	{0x0EF8, 0x0E10, 0x0D60, 0x0C80, 0x0BD8, 0x0B28, 0x0A88, 0x09F0, 0x0960, 0x08E0, 0x0858, 0x07E0},	// table 1
	{0x0D10, 0x0C55, 0x0BA4, 0x0AFC, 0x0A5F, 0x09CA, 0x093D, 0x08B8, 0x083B, 0x07C5, 0x0755, 0x06EC},	// table 2
	{0x0CDA, 0x0C22, 0x0B73, 0x0ACF, 0x0A33, 0x09A1, 0x0917, 0x0894, 0x0819, 0x07A4, 0x0737, 0x06CF}	// table 3
};

// very important(!) global pmd pos variable
volatile size_t pmd_pos = 0;

// pt3 pointers
unsigned char *pt3file;
pt3_hdr *pt3hdr;

// pmd headers
pt3md_hdr *pmdhdr;
pt3md_mdl_hdr *pmdmdl;

// global data
pmdins_t pmdins[32] = {0};
pmdorn_t pmdorn[16] = {0};

pmdpat_t pt3pat[MAX_PATTERNS] = {0};
std::vector<pmdpat_t> pmdpat;

u8 pt3ord[256] = {0};
std::vector<u8> pmdord;

size_t ordxlat[MAX_PATTERNS];

// optimized data table
std::vector<hdat> optpat, optasd;

// flags
bool autospd = false, verbose = false;

// -------------------------------------------------

// remove trailing right spaces
void rtrim(char *s) {
	char *se = s + (strlen(s) - 1);
	while(se >= s) {
		if (*se == ' ') *se-- = '\0'; else break; 
	}
}

char name[33] = {0}, author[33] = {0};

// make pmd header
int make_hdr(FILE *f) {
	wprintf(L"\nget header info...\n");

	// copy strings
	strncpy_s(name,   sizeof(char) * 33, pt3hdr->name, 32);   name[32]   = '\0';
	strncpy_s(author, sizeof(char) * 33, pt3hdr->author, 32); author[32] = '\0';

	// remove trailing spaces
	rtrim(name); rtrim(author); 

	wprintf(L"module: %S by %S \n", name, author);

	// prepare header
	memcpy(pmdhdr->sign, PT3MD_SIGN, 4);
	pmdhdr->ver.major = PT3MD_VER_MAJOR;
	pmdhdr->ver.minor = PT3MD_VER_MINOR;
	
	pmdhdr->type.chipnum  = 0;	// pt3-TS modules are not supported yet
	pmdhdr->type.mch_mode = pmdhdr_mch_mode::sync;

	//   
	pmdmdl->ord_num   = pt3hdr->song_end;
	pmdmdl->loop_pos  = pt3hdr->song_loop;
	pmdmdl->speed     = pt3hdr->speed;

	pmdmdl->int_rate  = (int)(15625 / PT3MD_INT_RATE);
	pmdmdl->clock     = PT3MD_CLOCK / 1000;

	pmdmdl->p_name    = pmd_pos = sizeof(pt3md_hdr) + sizeof(pt3md_mdl_hdr); pmd_pos += sizeof(char)*(strlen(name) + 1);
	pmdmdl->p_author  = pmd_pos; pmd_pos += sizeof(char)*(strlen(author) + 1);
	pmdmdl->p_notetab = pmd_pos; pmd_pos += sizeof(u16)*12;

	// container header is ready, write it
	fwrite(pmdhdr, sizeof(pt3md_hdr), 1, f);

	wprintf(L"done\n\n");

	return 0;
}

int get_orders(FILE *f) {
	wprintf(L"get order table: ");
	
	for (pmdmdl->ord_num = 0; (pmdmdl->ord_num < pt3hdr->song_end) || (pt3file[sizeof(pt3_hdr) + pmdmdl->ord_num] != 0xFF); pmdmdl->ord_num++) {
		pt3ord[pmdmdl->ord_num] = pt3file[sizeof(pt3_hdr) + pmdmdl->ord_num] / 3;
		pt3pat[pt3ord[pmdmdl->ord_num]].avail = true;
		wprintf(L".");
	}

	wprintf(L" %d positions -> done\n", pmdmdl->ord_num);
	return 0;
}

// convert instruments
int convert_ins(FILE *f) {
	wprintf(L"convert instruments: ");

	for (int i = 0; i < 32; i++) {
		pt3md_ins_hdr v = {0};

		// skip empty instrument
		if (!pt3hdr->instr[i]) {
			continue;
		}

		// else convert it
		wprintf(L".");

		unsigned char *p = &pt3file[pt3hdr->instr[i]];
		
		pt3_ins_hdr *h = (pt3_ins_hdr*)p;

		// push header data
		v.len        = h->len;
		v.loop_start = h->loop;
		v.linesize   = 4;
		v.sustain    = 0;

		pt3_ins_data *d = (pt3_ins_data*)(p + sizeof(pt3_ins_hdr));
		pt3md_ins_data t;

		//      
		for (int j = 0; j < h->len; j++) {
			if ((!d[j].e_mask) && (d[j].n_mask) && (d[j].n_ofs)) {
				v.linesize = 5;
				break;
			}
		}

		pmdins[i].avail = true;
		pmdins[i].hdr = v;

		//   ^W
		for (int j = 0; j < h->len; j++) {
			memset(&t, 0, sizeof(pt3md_ins_data));

			t.t_ofsl  = (d->t_mask? 0 : (u8)d->t_ofs);

			t.t_ofsh  = (d->t_mask? 0 : (u8)(d->t_ofs >> 8));
			t.t_mask  = d->t_mask;
			t.t_acc   = d->t_acc;
			t.n_acc   = d->n_acc;
			t.n_mask  = d->n_mask;

			t.n_ofs   = (d->n_mask ? 0 : d->n_ofs);
			t.v_slide = d->v_slide;

			t.volume  = d->volume;
			t.e_mask  = ~d->e_mask; //     AY (  pt3 )
			t.e_acc   = (((!d->e_mask) && (d->n_mask)) ? d->n_acc : 0);

			t.e_ofs   = (((!d->e_mask) && (d->n_mask)) ? d->n_ofs : 0);

			//   vector (KIDS DO NOT TRY THIS AT HOME!)
			u8 *hui = (u8*)&t;
			for (int k = 0; k < v.linesize; k++) {
				pmdins[i].data.push_back(hui[k]);
			}

			d++;
		}
	}

	//  - 
	for (pmdmdl->ins_num = 32; pmdmdl->ins_num > 0; pmdmdl->ins_num--) {
		if (pmdins[pmdmdl->ins_num-1].avail) break; 
	}

	wprintf(L" %d instruments -> done\n", pmdmdl->ins_num); 

	return 0;
}

int convert_orn(FILE *f) {
	wprintf(L"convert ornaments: ");
	
	for (int i = 0; i < 16; i++) {
		pt3md_orn v = {1, 0};

		// skip empty ornament
		if (!pt3hdr->orn[i]) {
			continue;
		}

		// else convert it
		wprintf(L".");

		unsigned char *p = &pt3file[pt3hdr->orn[i]];
		
		pt3_ins_hdr *h = (pt3_ins_hdr*)p;

		// push header data
		v.len        = h->len;
		v.loop_start = h->loop;

		pmdorn[i].avail = true;
		pmdorn[i].hdr = v;

		i8 *d = (i8*)(p + sizeof(pt3_ins_hdr));

		//    
		for (int j = 0; j < h->len; j++) pmdorn[i].data.push_back(*d++);
	};

	//  - 
	for (pmdmdl->orn_num = 16; pmdmdl->orn_num > 0; pmdmdl->orn_num--) {
		if (pmdorn[pmdmdl->orn_num-1].avail) break;
	}

	wprintf(L" %d ornaments -> done\n", pmdmdl->orn_num); 

	return 0;
}

#define convert_pat_warn(s) fwprintf(stderr, L"(%d, %d, %d) warning: %ls\n", k, patlen, i, s);

// convert patterns (ROCK 'N' ROLL!)
int convert_pat(FILE *f) {

	wprintf(L"convert patterns: ");

	for (int k = 0; k < MAX_PATTERNS; k++) {
		if (!pt3pat[k].avail) {
			pt3pat[k].hdr.patlen = pt3pat[k].hdr.autospd.len = 0; pt3pat[k].hdr.autospd.ofs = 0;
			pt3pat[k].pchan.p_chan_a = pt3pat[k].pchan.p_chan_b = pt3pat[k].pchan.p_chan_c = pt3pat[k].pchan.p_chan_d = 0;
			continue;
		}

		pt3_pattern_data pt3dat = {0};
		pt3md_pattern_data pmddat = {0};

		std::vector<pt3md_pattern_data> pmdstream;
		std::vector<int> asdpat;	// for autospeed

		// else start to parse

		unsigned char *chan[3];
		chan[0] = &pt3file[(unsigned short)(pt3file[(pt3hdr->pat_table + (k * 6)    )] + (pt3file[(pt3hdr->pat_table + (k * 6) + 1)] << 8))];
		chan[1] = &pt3file[(unsigned short)(pt3file[(pt3hdr->pat_table + (k * 6) + 2)] + (pt3file[(pt3hdr->pat_table + (k * 6) + 3)] << 8))];
		chan[2] = &pt3file[(unsigned short)(pt3file[(pt3hdr->pat_table + (k * 6) + 4)] + (pt3file[(pt3hdr->pat_table + (k * 6) + 5)] << 8))];

		// 
		int skipstart[3] = {0}, skip[3] = {0}, eop[3] = {false};
		size_t patlen = 0;
		int prevvol[3] = {-1, -1, -1};
		int prevenv[3] = {0};

		// paaaaaaarse patterns.....
		while((!eop[0]) && (!eop[1]) && (!eop[2])) {

			// clear pt3dat
			pt3dat.env = -1; pt3dat.noise = (patlen ? -1 : 0); // well-known pt3 feature
			for (int i = 0; i < 3; i++) {
				pt3dat.chan[i].cmd = pt3dat.chan[i].env = pt3dat.chan[i].ins = \
				pt3dat.chan[i].note = pt3dat.chan[i].orn = pt3dat.chan[i].parm = \
				pt3dat.chan[i].subparm = pt3dat.chan[i].vol = -1;
			}

			for (int i = 0; i < 3; i++) {
				if (eop[i]) continue;
				if (skip[i]) {skip[i]--; continue;} else skip[i] = skipstart[i];

				int stop = false, iscmd = false, cmd = 0;
				while (!stop) {
					// get one event and process it
					unsigned char ch = *chan[i]++;

					// command
					// warning: only ONE command per line is supported!
					if ((ch >= 0x01) && (ch <= 0x09)) {
						iscmd = true;
						cmd = ch;
					} 
					
					// set instrument, 0th ornament and new env period
					if ((ch >= 0x11) && (ch <= 0x1F)) {
						pt3dat.chan[i].env = (ch & 0xF);		// ptdoc.txt errata: (ch & 0xF) - 1
						pt3dat.env  = (*chan[i]++ << 8);
						pt3dat.env +=  *chan[i]++;
						pt3dat.chan[i].orn = 0;
						pt3dat.chan[i].ins = (*chan[i]++ >> 1);
					} 

					// set noise
					if ((ch >= 0x20) && (ch <= 0x3F)) {
						pt3dat.noise = (ch & 0x1F);
						if (i != 1) convert_pat_warn(L"set noise event not in channel B");
					}

					// set ornament
					if ((ch >= 0x40) && (ch <= 0x4F)) pt3dat.chan[i].orn = (ch & 0xF);

					// set note and end of line
					if ((ch >= 0x50) && (ch <= 0xAF)) {
						pt3dat.chan[i].note = (ch - 0x50);
						stop = true;
					}

					// set new env period
					if ((ch >= 0xB2) && (ch <= 0xBF)) {
						pt3dat.chan[i].env = (ch & 0xF) - 1;
						pt3dat.env  = (*chan[i]++ << 8);
						pt3dat.env +=  *chan[i]++;
					}

					// set volume
					if ((ch >= 0xC1) && (ch <= 0xCF)) {
						pt3dat.chan[i].vol = (ch & 0xF);
					}

					// set instrument
					if ((ch >= 0xD1) && (ch <= 0xEF)) pt3dat.chan[i].ins = (ch - 0xD0);

					// disable envelope, set instument and ornament
					if ((ch >= 0xF0) && (ch <= 0xFF)) {
						pt3dat.chan[i].env = 0xF;
						pt3dat.chan[i].orn = (ch & 0xF);
						pt3dat.chan[i].ins = (*chan[i]++ >> 1);
					}

					switch(ch) {
						// end of channel
					case 0x00:
						stop   = true;
						eop[i] = true;
						break;

						// disable envelope, set instrument and 0th ornament
					case 0x10:
						pt3dat.chan[i].env = 0xF;
						pt3dat.chan[i].ins = (*chan[i]++ >> 1);
						break;

						// disable envelope
					case 0xB0:
						pt3dat.chan[i].env = 0xF;
						break;
						
						// delay
					case 0xB1:
						skipstart[i] = (*chan[i]++) - 1;
						skip[i] = skipstart[i];
						break;

						// R-- + end of line
					case 0xC0:
						pt3dat.chan[i].note = -2;
						stop = true;
						break;

						// end of line
					case 0xD0:
						stop = true;
						break;
					}
				}

				// check for commands
				if (iscmd) {
					signed short rawslide;

					switch(cmd) {
					case 0x1:
						// cmd 1/2 - tone period slide
						pt3dat.chan[i].subparm = *chan[i]++;
						rawslide  =  (*chan[i] + (*(chan[i] + 1) << 8));
						if (rawslide > 0) {
							pt3dat.chan[i].cmd  = 1;
							pt3dat.chan[i].parm = rawslide & 0xFF;
						} else {
							pt3dat.chan[i].cmd  = 2;
							pt3dat.chan[i].parm = (-rawslide) & 0xFF;
						}
						chan[i] += 2;
						break;

					case 0x2:
						// cmd 3 - tone portamento
						pt3dat.chan[i].subparm = *chan[i]++; chan[i] += 2;
						rawslide  =  (*chan[i] + (*(chan[i] + 1) << 8));
						pt3dat.chan[i].cmd  = 3;
						pt3dat.chan[i].parm = abs(rawslide) & 0xFF;
						chan[i] += 2;
						break;

					case 0x3:
						// cmd 4 - instrument offset
						pt3dat.chan[i].cmd  = 4;
						pt3dat.chan[i].parm = *chan[i]++;
					    break;

					case 0x4:
						// cmd 5 - ornament offset
						pt3dat.chan[i].cmd  = 5;
						pt3dat.chan[i].parm = *chan[i]++;
						break;

					case 0x5:
						// cmd 6 - tremor
						pt3dat.chan[i].cmd  = 6;
						pt3dat.chan[i].parm = ((*chan[i] & 0xF) << 4) + (*(chan[i] + 1) & 0xF);
						chan[i] += 2;
						break;

					case 0x8:
						// cmd 9/A - envelope period slide
						pt3dat.chan[i].subparm = *chan[i]++;
						rawslide  =  (*chan[i] + (*(chan[i] + 1) << 8));
						if (rawslide > 0) {
							pt3dat.chan[i].cmd  = 9;
							pt3dat.chan[i].parm = rawslide & 0xFF;
						} else {
							pt3dat.chan[i].cmd  = 0xA;
							pt3dat.chan[i].parm = (-rawslide) & 0xFF;
						}
						chan[i] += 2;
						
						break;

					case 0x9:
						// cmd B - set speed
						pt3dat.chan[i].cmd  = 0xB;
						pt3dat.chan[i].parm = *chan[i]++;
						break;

					default:
						break;
					}
				}
			}

			//       pt3 -      pmd
			pmddat.env_chan.period = pt3dat.env;
			pmddat.env_chan.noise = pt3dat.noise;

			pmddat.env_chan.orn = -1;	//    pt3
			pmddat.env_chan.type = -1;

			pmddat.env_chan.cmd = pmddat.env_chan.subparm = pmddat.env_chan.parm = -1;

			int newspd = 0;	// new speed

			for (int i = 0; i < 3; i++) {
				//   
				if (pt3dat.chan[i].env != -1) {
					if (pt3dat.chan[i].env != 0xF) {
						if ((pt3dat.chan[i].env >= 0) && (pt3dat.chan[i].env <= 3)) pmddat.env_chan.type = 9; else
						if ((pt3dat.chan[i].env >= 4) && (pt3dat.chan[i].env <= 7)) pmddat.env_chan.type = 0xF; else
							pmddat.env_chan.type = pt3dat.chan[i].env & 0x7;
						pmddat.chan[i].env = 1;
						if ((pmddat.env_chan.cmd == 1) || (pmddat.env_chan.cmd == 2)) pmddat.env_chan.cmd = 0;  
					} else pmddat.chan[i].env = 0;  //  0xF  pt3 -   ;)
				} else pmddat.chan[i].env = -1; 

				//    -  ..
				pmddat.chan[i].note = pt3dat.chan[i].note;
				pmddat.chan[i].ins = pt3dat.chan[i].ins;
				pmddat.chan[i].orn = pt3dat.chan[i].orn;
				
				if (pt3dat.chan[i].vol != prevvol[i]) {
					pmddat.chan[i].vol = pt3dat.chan[i].vol;
					prevvol[i]         = pmddat.chan[i].vol;
				} else pmddat.chan[i].vol = -1;

				if ((pt3dat.chan[i].cmd != 0x9) && (pt3dat.chan[i].cmd != 0xA)) {
					pmddat.chan[i].cmd = pt3dat.chan[i].cmd;
					pmddat.chan[i].subparm = pt3dat.chan[i].subparm;
					pmddat.chan[i].parm = pt3dat.chan[i].parm;
				} else {
					pmddat.env_chan.cmd = (pt3dat.chan[i].cmd - 0x8);
					pmddat.env_chan.subparm = pt3dat.chan[i].subparm;
					pmddat.env_chan.parm = pt3dat.chan[i].parm;

					pmddat.chan[i].cmd = pmddat.chan[i].subparm = pmddat.chan[i].parm = -1;
				}

				if (pmddat.chan[i].cmd == 0xB) 
					if (autospd) {
						newspd = pmddat.chan[i].parm;
						pmddat.chan[i].cmd = pmddat.chan[i].subparm = pmddat.chan[i].parm = -1;
					} else {
						pmddat.chan[i].cmd = 0xF; pmddat.chan[i].subparm = 0;
					}
			}

			//  
			pmdstream.push_back(pmddat);
			asdpat.push_back(newspd);

			patlen++;
		}
		
		//  
		// patlen     ,   pmdstream   
		patlen--; pmdstream.pop_back(); asdpat.pop_back();

		if (autospd) {
			std::vector<int> r = asd_findpattern(asdpat);
			for (size_t i = 0; i < r.size(); i++) pt3pat[k].autospd.p.push_back((u8)r[i]); 
		}

		//    ,   ^_^
		//   
		pt3pat[k].avail = true;
		pt3pat[k].hdr.patlen = patlen;
		pt3pat[k].hdr.autospd.len = (autospd ? ((pt3pat[k].autospd.p.size() > 1) ? pt3pat[k].autospd.p.size() : 0) : 0);
		pt3pat[k].hdr.autospd.ofs = 0;

		//   
		if (autospd) { 
			if (pt3pat[k].autospd.p.size() <= 1) {
				// restore set speed commands
				for (size_t j = 0; j < patlen; j++) if (asdpat[j]) spd_push(asdpat[j], &pmdstream[j]);
			} else {
				bool nasd = false;	// don't use autospeed
				bool rasd = true;	// repeat autospd
				int  nacc = 0;

				for (size_t i = 0; i < patlen; i++) {
					if (asdpat[i] == pt3pat[k].autospd.p[i % pt3pat[k].hdr.autospd.len]) {
						if (!rasd) spd_push(0, &pmdstream[i]);
						rasd = true;
					} else  {
						rasd = false;
						if (asdpat[i]) {
							spd_push(asdpat[i], &pmdstream[i]);
							nacc = asdpat[i];
						} else {
							size_t k = i;
							while ((k >= 0) && (!asdpat[k])) k--;
							if (asdpat[k] != nacc) {
								nacc = asdpat[k];
								spd_push(asdpat[i], &pmdstream[i]);
							}
						}
					}
				}
			}
		}

		//     ,   ,   
		//    
		struct {
			struct chan {
				pmd_chan_event_NN nn;
				pmd_chan_event_XV xv;
				pmd_chan_event_II ii;
				pmd_chan_event_OO oo;
				pmd_chan_event_DC dc;
				pmd_chan_event_EE ee;
			} chan[3];
			pmd_env_event_XT xt;
			pmd_env_event_PP pp;
			pmd_env_event_OO oo;
			pmd_env_event_NN nn;
			pmd_env_event_DC dc;
			pmd_env_event_EE ee;
		} pmdevent;

		size_t skipenv = 0;
		for (size_t j = 0; j < patlen; j++) {
			bool empty = false; size_t emptypos = 0;
			
			//    
			for (emptypos = j+1; emptypos < patlen; emptypos++) {
				//   
				pmddat = pmdstream.at(emptypos);

				//    
				if (!((pmddat.env_chan.type == -1) && (pmddat.env_chan.period == -1) && (pmddat.env_chan.noise == -1) &&
					  (pmddat.env_chan.cmd == -1))) break;
			}
			
			emptypos -= j;

			if (skipenv != emptypos) {
				skipenv = emptypos;
				pt3pat[k].data[3].p.push_back(emptypos > 7 ? 0 : emptypos);
				if (emptypos > 7) pt3pat[k].data[3].p.push_back(emptypos);
			};

			//   
			pmddat = pmdstream.at(j);

			//   \
			pmdevent.xt.type    = (pmddat.env_chan.type != -1) ? pmddat.env_chan.type : 0;
			pmdevent.xt.restart = (pmddat.env_chan.type != -1) ? 1 : 0;	// pt3      
			if (pmddat.env_chan.period != -1) 
				pmdevent.xt.fmt = (pmddat.env_chan.period >= 0x100 ? pmd_env_event_XT_fmt::period : pmd_env_event_XT_fmt::period_lsb);
			else pmdevent.xt.fmt = pmd_env_event_XT_fmt::none;

			pmdevent.xt.noise   = (pmddat.env_chan.noise != -1) ? 1 : 0;
			pmdevent.xt.cmd     = (pmddat.env_chan.cmd   != -1) ? 1 : 0;

			//   ( )
			pmdevent.pp.period_lsb = (pmdevent.xt.fmt != pmd_env_event_XT_fmt::none)   ? (pmddat.env_chan.period & 0xFF) : 0;
			pmdevent.oo            = (pmdevent.xt.fmt == pmd_env_event_XT_fmt::period) ? (pmddat.env_chan.period >> 8)   : 0;

			//   ( )
			pmdevent.nn            = (pmddat.env_chan.noise != -1) ? pmddat.env_chan.noise : 0;

			//   ( )
			pmdevent.dc.cmd        = (pmddat.env_chan.cmd   != -1) ? pmddat.env_chan.cmd : 0;
			pmdevent.dc.subparm    = (pmddat.env_chan.cmd   != -1) ? (pmddat.env_chan.subparm == -1 ? 0 : pmddat.env_chan.subparm) : 0;
			pmdevent.ee            = (pmddat.env_chan.cmd   != -1) ? pmddat.env_chan.parm : 0;

			//  
			//  XT
			pt3pat[k].data[3].p.push_back(*((u8*)(&pmdevent.xt)));

			//  \\whatever
			switch(pmdevent.xt.fmt) {
			case pmd_env_event_XT_fmt::note:
				pt3pat[k].data[3].p.push_back(*((u8*)(&pmdevent.pp)));
				if (pmdevent.pp.orn) pt3pat[k].data[3].p.push_back(*((u8*)(&pmdevent.oo)));
				break;
			case pmd_env_event_XT_fmt::period_lsb:
				pt3pat[k].data[3].p.push_back(*((u8*)(&pmdevent.pp)));
				break;
			case pmd_env_event_XT_fmt::period:
				pt3pat[k].data[3].p.push_back(*((u8*)(&pmdevent.pp)));
				pt3pat[k].data[3].p.push_back(*((u8*)(&pmdevent.oo)));
				break;
			default:
				break;
			}

			//   ( )
			if (pmdevent.xt.noise)   pt3pat[k].data[3].p.push_back(*((u8*)(&pmdevent.nn)));

			//   ( )
			if (pmdevent.xt.cmd) switch(pmdevent.dc.cmd) {
			case 0:
				break;

			case 0x5:
			case 0xF:
				pt3pat[k].data[3].p.push_back(*((u8*)(&pmdevent.dc)));
				pt3pat[k].data[3].p.push_back(*((u8*)(&pmdevent.ee)));
				break;

			default:
				pt3pat[k].data[3].p.push_back(*((u8*)(&pmdevent.dc)));
				if (pmdevent.dc.subparm) pt3pat[k].data[3].p.push_back(*((u8*)(&pmdevent.ee)));
				break;
			}

			j += skipenv-1;
		}

		for (int i = 0; i < 3; i++) {
			volatile unsigned int oldvol = 0, oldenv = 0;
			volatile bool firstvol = true, firstenv = true;

			volatile size_t skipcnt = 0;
			for (size_t j = 0; j < patlen; j++) {
				bool empty = false; size_t emptypos = 0;
				
				//    
				for (emptypos = j+1; emptypos < patlen; emptypos++) {
					//   
					pmddat = pmdstream.at(emptypos);

					bool nochange = ((pmddat.chan[i].vol == -1)) && ((pmddat.chan[i].env == -1));

					//    
					if (!((nochange) && (pmddat.chan[i].note == -1) && (pmddat.chan[i].cmd == -1) &&
						  (pmddat.chan[i].ins == -1) && (pmddat.chan[i].orn == -1)))  break;
				}

				emptypos -= j;

				if (skipcnt != emptypos) {
					skipcnt = emptypos;
					pt3pat[k].data[i].p.push_back(emptypos > 8 ? 0x67 : emptypos + 0x67);
					if (emptypos > 8) pt3pat[k].data[i].p.push_back(emptypos);
				};
				
				//   
				pmddat = pmdstream.at(j);
				
				//   A-C


				//    
				if ((firstvol) && (pmddat.chan[i].vol != -1)) {
					pt3pat[k].data[i].p.push_back(0xF0 | pmddat.chan[i].vol);
					oldvol = pmddat.chan[i].vol;
					firstvol = false;
				}

				if ((firstenv) && (pmddat.chan[i].env != -1)) {
					pt3pat[k].data[i].p.push_back(pmddat.chan[i].env ? 0x65 : 0x64);
					oldenv = pmddat.chan[i].env;
					firstenv = false;
				}

				//  
				pmdevent.chan[i].nn.note = (pmddat.chan[i].note != -1) ? (pmddat.chan[i].note == -2 ? pmd_chan_event_NN_note::notecut : pmddat.chan[i].note + 1) : 0;

				//   XV
				bool nochange = ((pmddat.chan[i].vol == -1) || (pmddat.chan[i].vol == oldvol)) && ((pmddat.chan[i].env == -1) || (pmddat.chan[i].env == oldenv));

				if ((!nochange) || (pmddat.chan[i].cmd != -1) || (pmddat.chan[i].ins != -1) ||
					(pmddat.chan[i].orn != -1)) pmdevent.chan[i].nn.notlast = 1;
				else pmdevent.chan[i].nn.notlast = 0;

				//  \ ( )
				pmdevent.chan[i].xv.vol  = (pmddat.chan[i].vol != -1) ? oldvol ^ pmddat.chan[i].vol : 0;
				pmdevent.chan[i].xv.env  = (pmddat.chan[i].env != -1) ? oldenv ^ pmddat.chan[i].env : 0;
				pmdevent.chan[i].xv.ins  = (pmddat.chan[i].ins != -1) ? 1 : 0;
				pmdevent.chan[i].xv.orn  = (pmddat.chan[i].orn != -1) ? 1 : 0;
				pmdevent.chan[i].xv.cmd  = (pmddat.chan[i].cmd != -1) ? 1 : 0;
				
				//   
				if ((pmddat.chan[i].vol) != -1) oldvol = pmddat.chan[i].vol;
				if ((pmddat.chan[i].env) != -1) oldenv = pmddat.chan[i].env;

				//   ( )
				pmdevent.chan[i].ii      = (pmddat.chan[i].ins != -1) ? pmddat.chan[i].ins : 0;

				//   ( )
				pmdevent.chan[i].oo      = (pmddat.chan[i].orn != -1) ? pmddat.chan[i].orn : 0;

				//   ( )
				pmdevent.chan[i].dc.cmd     = (pmddat.chan[i].cmd   != -1) ? pmddat.chan[i].cmd : 0;
				pmdevent.chan[i].dc.subparm = (pmddat.chan[i].cmd   != -1) ? (pmddat.chan[i].subparm == -1 ? 0 : pmddat.chan[i].subparm) : 0;
				pmdevent.chan[i].ee         = (pmddat.chan[i].cmd   != -1) ? pmddat.chan[i].parm : 0;

				// "HERE I COME!" -   ,      
				//        =) ( ?)

				//  
				pt3pat[k].data[i].p.push_back(*((u8*)(&pmdevent.chan[i].nn)));

				//  
				if (pmdevent.chan[i].nn.notlast) {
					pt3pat[k].data[i].p.push_back(*((u8*)(&pmdevent.chan[i].xv)));

					if (pmdevent.chan[i].xv.ins) pt3pat[k].data[i].p.push_back(*((u8*)(&pmdevent.chan[i].ii)));
					if (pmdevent.chan[i].xv.orn) pt3pat[k].data[i].p.push_back(*((u8*)(&pmdevent.chan[i].oo)));

					if (pmdevent.chan[i].xv.cmd) switch(pmdevent.chan[i].dc.cmd) {
					case 0:
						break;

					case 0x4:
					case 0x5:
					case 0x6:
					case 0xF:
						pt3pat[k].data[i].p.push_back(*((u8*)(&pmdevent.chan[i].dc)));
						pt3pat[k].data[i].p.push_back(*((u8*)(&pmdevent.chan[i].ee)));
						break;

					default:
						pt3pat[k].data[i].p.push_back(*((u8*)(&pmdevent.chan[i].dc)));
						if (pmdevent.chan[i].dc.subparm) pt3pat[k].data[i].p.push_back(*((u8*)(&pmdevent.chan[i].ee)));
						break;
					}

				}

				j += skipcnt-1;
			}
		}

		//    
		for (size_t i = 0; i < 4; i++) {
			pt3pat[k].data[i].hash = XXH32(pt3pat[k].data[i].p.data(), sizeof(u8)*pt3pat[k].data[i].p.size(), 0);
		}
		pt3pat[k].autospd.hash = XXH32(pt3pat[k].autospd.p.data(), sizeof(u8)*pt3pat[k].autospd.p.size(), 0);

		wprintf(L".");
		pmdmdl->pat_num++;
	}

	//      
	size_t rpnum = 0; pmdmdl->pat_num = 0;
	for (size_t i = 0; i < MAX_PATTERNS; i++) {
		if (pt3pat[i].avail) {
			pmdpat.push_back(pt3pat[i]);
			ordxlat[i] = rpnum++;
			pmdmdl->pat_num++;
		} else ordxlat[i] = -1;
	}

	//   
	for (size_t i = 0; i < pmdmdl->ord_num; i++) {
		if (ordxlat[pt3ord[i]] == -1) {
			// this should never occur btw
			fprintf(stderr, "\nerror: ord %d - incorrect pattern number!\n", i);
			return 1;
		} else pmdord.push_back(ordxlat[pt3ord[i]]);
	}

	wprintf(L" %d patterns -> done\n", pmdmdl->pat_num); 

	return 0;
}

int write_pmd(FILE *f) {
	u16 *insaddr, *ornaddr;

	if (pmdmdl->ins_num) {
		insaddr = new u16[pmdmdl->ins_num];
		
		//  pmd_pos        , 
		//   
		pmdmdl->p_instab = pmd_pos; pmd_pos += sizeof(u16)*pmdmdl->ins_num;
	
		for (int i = 0; i < pmdmdl->ins_num; i++) {
			if (pmdins[i].avail) {
				insaddr[i] = pmd_pos; pmd_pos += (sizeof(pmdins[i].hdr) + pmdins[i].data.size());
			} else
				insaddr[i] = 0;
		}
	}

	if (pmdmdl->orn_num) {
		ornaddr = new u16[pmdmdl->orn_num];
		
		//  pmd_pos        , 
		//   
		pmdmdl->p_orntab = pmd_pos; pmd_pos += sizeof(u16)*pmdmdl->orn_num;
	
		for (int i = 0; i < pmdmdl->orn_num; i++) {
			if (pmdorn[i].avail) {
				ornaddr[i] = pmd_pos; pmd_pos += (sizeof(pmdorn[i].hdr) + pmdorn[i].data.size());
			} else
				ornaddr[i] = 0;
		}
	}

	if (pmdmdl->ord_num) {
		pmdmdl->p_order = pmd_pos; pmd_pos += (sizeof(u8) * pmdmdl->ord_num);
	}

	if (pmdmdl->pat_num) {
		// reserve memory for pattern table
		pmdmdl->p_pattab = pmd_pos; pmd_pos += (sizeof(pt3md_pattern_pchan) + sizeof(pt3md_pattern)) * pmdmdl->pat_num;

		//    optidx
		for (size_t i = 0; i < pmdmdl->pat_num; i++) {
			if (pmdpat[i].avail) { 
				pmdpat[i].pchan.p_chan_a = pmd_pos + optpat[pmdpat[i].data[0].optidx].optidx;
				pmdpat[i].pchan.p_chan_b = pmd_pos + optpat[pmdpat[i].data[1].optidx].optidx;
				pmdpat[i].pchan.p_chan_c = pmd_pos + optpat[pmdpat[i].data[2].optidx].optidx;
				pmdpat[i].pchan.p_chan_d = pmd_pos + optpat[pmdpat[i].data[3].optidx].optidx;

			}
		}
		pmd_pos += optpos;

		if (autospd) {
			//    
			for (size_t i = 0; i < pmdmdl->pat_num; i++) {
				if (pmdpat[i].avail)
					pmdpat[i].hdr.autospd.ofs = pmd_pos + optasd[pmdpat[i].autospd.optidx].optidx;
			}
			pmd_pos += asdpos;
		}
	}


	//     :)
	fwrite(pmdmdl,  sizeof(pt3md_mdl_hdr), 1, f);
	fwrite(name,    sizeof(char)*(strlen(name)   + 1), 1, f);
	fwrite(author,  sizeof(char)*(strlen(author) + 1), 1, f);
	
	fwrite(&pt3_tonetab[pt3hdr->tonetable], sizeof(u16)*12, 1, f);

	if (pmdmdl->ins_num) {
		fwrite(insaddr, sizeof(u16)*pmdmdl->ins_num, 1, f);
		for (int i = 0; i < pmdmdl->ins_num; i++) {
			if (insaddr[i]) {
				fwrite(&pmdins[i].hdr,        sizeof(pmdins[i].hdr), 1, f);
				fwrite(pmdins[i].data.data(), pmdins[i].data.size(), 1, f);
			}
		}
	}

	if (pmdmdl->orn_num) {
		fwrite(ornaddr, sizeof(u16)*pmdmdl->orn_num, 1, f);
		for (int i = 0; i < pmdmdl->orn_num; i++) {
			if (ornaddr[i]) {
				fwrite(&pmdorn[i].hdr,        sizeof(pmdorn[i].hdr), 1, f);
				fwrite(pmdorn[i].data.data(), pmdorn[i].data.size(), 1, f);
			}
		}
	}

	if (pmdmdl->ord_num) fwrite(pmdord.data(), sizeof(u8) * pmdmdl->ord_num, 1, f);

	if (pmdmdl->pat_num) {
		for (size_t i = 0; i < pmdmdl->pat_num; i++) {
			fwrite(&pmdpat[i].pchan, sizeof(pt3md_pattern_pchan), 1, f);
			fwrite(&pmdpat[i].hdr,   sizeof(pt3md_pattern), 1, f);
		}
		for (size_t i = 0; i < optpat.size(); i++) fwrite(optpat[i].p.data(), sizeof(u8)*optpat[i].p.size(), 1, f);
		if (autospd) for (size_t i = 0; i < optasd.size(); i++) fwrite(optasd[i].p.data(), sizeof(u8)*optasd[i].p.size(), 1, f);
	}

	// free memory
	if (pmdmdl->ins_num) delete[] insaddr; 
	if (pmdmdl->orn_num) delete[] ornaddr;

	return 0;
}

int _tmain(int argc, _TCHAR* argv[])
{
	_TCHAR pt3name[256], pt3mdname[256];
	FILE *pt3, *pt3md;
	size_t parm_arg = 3;
	
	wprintf(L"Pro Tracker 3.x -> pt3md module converter - by wbcbz7 z8.o5.zol7\n");

    if (argc < 2) {
		fwprintf(stderr, L"usage: pmdconv.exe [filename.pt3] <filename.pmd> <arg>...\n");
		fwprintf(stderr, L"\n");
		fwprintf(stderr, L" --autospd - use autospeed (experimental)\n");
		fwprintf(stderr, L"\n");
		fwprintf(stderr, L" --clk     - chip clock rate in hz\n");
		fwprintf(stderr, L" --int     - int rate in hz\n");
		fwprintf(stderr, L" --stereo  - stereo layout (abc/acb/bac/mono)\n");
		fwprintf(stderr, L" --chip    - chip type (ay/ym)\n");
		fwprintf(stderr, L"\n");
		fwprintf(stderr, L"example: pmdconv.exe module.pt3 --clk=1750000 --int=48.8 --stereo=abc --chip=ay\n");
		return 1;
	}
	
	wcscpy_s(pt3name, 256, argv[1]);

    if ((argc < 3) || (argv[2][0] == L'-') || (!wcsstr(argv[2], L".") )) {
        _TCHAR *pp = wcsstr(argv[1], L".");
        wcsncpy_s(pt3mdname, 256, argv[1], (size_t)(pp - argv[1]));
        wcscat_s(pt3mdname, 256, L".pmd");
        parm_arg = 2;
    } else {
        wcscpy_s(pt3mdname, 256, argv[2]);
        parm_arg = 3;
    }

	_wfopen_s(&pt3, pt3name, L"rb");
	if (!pt3) {
		fwprintf(stderr, L"error opening file %ls: ", pt3name); perror(0);
		return 1;
	}

	// get filesize
	fseek(pt3, 0, SEEK_END);
	size_t pt3_fsize = ftell(pt3);

	// allocate maximum memory (also sandbox for broken modules)
	pt3file = new unsigned char[max(65536, pt3_fsize)];
	memset(pt3file, 0, sizeof(unsigned char) * max(65536, pt3_fsize));

	// read all file at once
	fseek(pt3, 0, SEEK_SET);
	fread(pt3file, sizeof(unsigned char) * pt3_fsize, 1, pt3);

	if (ferror(pt3)) {
		fwprintf(stderr, L"error reading file %ls: ", pt3name); perror(0);
		delete[] pt3file;
		return 1;
	}

	fclose(pt3);

	// parse file
	pt3hdr = (pt3_hdr*)pt3file;

	wprintf(L"\ninfile -> %ls - ", pt3name);
	// check file format
	if (!isalnum(pt3hdr->ver)) {
		fwprintf(stderr, L"error: incorrect file format!");
		delete[] pt3file;
		return 1;
	}
	if (isdigit(pt3hdr->ver)) wprintf(L"ProTracker 3.%c\n", pt3hdr->ver); else wprintf(L"Vortex Tracker\n");

	if (pt3hdr->ver < '6') fwprintf(stderr, L"warning: Pro Tracker 3.5 or older version module detected!\n");

	// allocate memory for pt3md headers
	pmdhdr = new pt3md_hdr; pmdmdl = new pt3md_mdl_hdr;
	memset(pmdhdr, 0, sizeof(pt3md_hdr));
	memset(pmdmdl, 0, sizeof(pt3md_mdl_hdr));

	// open pt3md file
	_wfopen_s(&pt3md, pt3mdname, L"wb");
	if (!pt3md) {
		fwprintf(stderr, L"error opening file %ls: ", pt3mdname); perror(0);
		return 1;
	}

	if (make_hdr(pt3md)) return 1;

	for (int i = parm_arg; i < argc; i++) {
        static _TCHAR arg[256];
		unsigned long tmp; float ftmp;
        wcscpy_s(arg, sizeof(_TCHAR)*256, argv[i]);
        _wcsupr_s(arg, sizeof(_TCHAR)*256);

		// toggle autospeed
		if (wcsstr(arg, L"A")) {
			autospd = true;
			continue;
        }

		// parse clock rate
		if (wcsstr(arg, L"CLK=")) {
			if (swscanf_s(wcsstr(arg, L"CLK=") + wcslen(L"CLK="), L"%d", &tmp)) pmdmdl->clock = (unsigned short)(tmp / 1000);
			continue;
        }

		// parse int rate
		if (wcsstr(arg, L"INT=")) {
			if (swscanf_s(wcsstr(arg, L"INT=") + wcslen(L"INT="), L"%f", &ftmp)) pmdmdl->int_rate = (unsigned short)(15625.0f / ftmp);
			else if (swscanf_s(wcsstr(arg, L"INT=") + wcslen(L"INT="), L"%d", &tmp)) pmdmdl->int_rate = (unsigned short)(15625 / tmp);
			continue;
        }

		// parse stereo layout
		if (wcsstr(arg, L"STEREO=")) {
			_TCHAR *p = wcsstr(arg, L"STEREO=") + wcslen(L"STEREO=");
			if (!wcsncmp(p, L"ABC",  wcslen(p))) pmdmdl->chip.stereo = pt3md_mdl_hdr_chip_stereo::abc;
			if (!wcsncmp(p, L"ACB",  wcslen(p))) pmdmdl->chip.stereo = pt3md_mdl_hdr_chip_stereo::acb;
			if (!wcsncmp(p, L"BAC",  wcslen(p))) pmdmdl->chip.stereo = pt3md_mdl_hdr_chip_stereo::bac;
			if (!wcsncmp(p, L"MONO", wcslen(p))) pmdmdl->chip.stereo = pt3md_mdl_hdr_chip_stereo::mono;
			continue;
        }

		// parse chip type
		if (wcsstr(arg, L"CHIP=")) {
			_TCHAR *p = wcsstr(arg, L"CHIP=") + wcslen(L"CHIP=");
			if (!wcsncmp(p, L"AY",  wcslen(p))) pmdmdl->chip.chiptype = pt3md_mdl_hdr_chip_type::ay;
			if (!wcsncmp(p, L"YM",  wcslen(p))) pmdmdl->chip.chiptype = pt3md_mdl_hdr_chip_type::ym;
			continue;
        }
	}

	wprintf(L"chip type:     %s\n", !pmdmdl->chip.chiptype ? L"AY" : L"YM");
	wprintf(L"stereo layout: ");
	switch (pmdmdl->chip.stereo) {
	case pt3md_mdl_hdr_chip_stereo::abc:  wprintf(L"ABC"); break;
	case pt3md_mdl_hdr_chip_stereo::acb:  wprintf(L"ACB"); break;
	case pt3md_mdl_hdr_chip_stereo::bac:  wprintf(L"BAC"); break;
	case pt3md_mdl_hdr_chip_stereo::mono: wprintf(L"mono"); break;
	default: wprintf(L"unknown"); break;
	}
	wprintf(L"\nclock rate:    %d Hz\n", pmdmdl->clock * 1000);
	wprintf(L"int rate:      %f Hz\n", (float)(15625.0f / pmdmdl->int_rate));

	if (get_orders(pt3md)) return 1;
	if (convert_ins(pt3md)) return 1;
	if (convert_orn(pt3md)) return 1;
	if (convert_pat(pt3md)) return 1;
	if (pat_optimize()) return 1;
	if (autospd) if (asd_optimize()) return 1;

	if (write_pmd(pt3md)) return 1;

	fclose(pt3md);

	wprintf(L"\noutfile -> %ls - pt3md %d.%d \n", pt3mdname, pmdhdr->ver.major, pmdhdr->ver.minor);
	wprintf(L"done\n");

	return 0;
}

