#pragma once

#include <vector>

// a really nice hash algo, check it out! :) https://github.com/Cyan4973/xxHash/
#include "xxhash\xxhash.h"

#define max(a, b) (a > b ? a : b)
#define min(a, b) (a < b ? a : b)

#define u8  unsigned char
#define i8  signed char
#define u16 unsigned short
#define i16 signed short

#define PT3MD_SIGN "PMD\x1A"

#define PT3MD_VER_MAJOR 0
#define PT3MD_VER_MINOR 3

#define PT3MD_INT_RATE  50
#define PT3MD_CLOCK     1750000

#define MAX_PATTERNS    256 / 3

enum pmdhdr_mch_mode {
	sync, async
};

// pt3md structs
#include "pt3md.h"

// pt3 structures

#pragma pack(push, 1)

typedef struct {
	char    id1[13];	// "ProTracker 3." or "Vortex Tracke"
	u8      ver;		// ASCII version number ('3'..'7') or 'r' for Vortex Tracker modules
	char    id2[16];	// " compilation of " or " II 1.0 module: "
	char    name[32];   // module name (ASCII)
	char    by[4];      // " by " or " By "
	char    author[32]; // author name (ASCII)
	u8      space;		// ASCII ' '
	u8      tonetable;  // tone table
	u8      speed;		// initial speed
	u8      song_end;	// number of positions
	u8      song_loop;  // loop position
	u16     pat_table;  // pattern table offset
	u16     instr[32];  // instrument data offset
	u16     orn[16];    // ornament data offset
} pt3_hdr;

typedef struct {
	u8  loop;
	u8  len;
} pt3_ins_hdr;

typedef struct {
	u8  e_mask:1;
	u8  n_ofs:5;
	u8  v_slide:2;

	u8  volume:4;
	u8  t_mask:1;
	u8  n_acc:1;
	u8  t_acc:1;
	u8  n_mask:1;

	i16 t_ofs;
} pt3_ins_data;

typedef struct {
	unsigned int env;
	unsigned int noise;
	struct chan_dat {
		unsigned int note;
		
		unsigned int ins;
		unsigned int orn;
		unsigned int env;
		unsigned int vol;

		unsigned int cmd;
		unsigned int subparm;
		unsigned int parm;
	} chan[3];
} pt3_pattern_data;


#define PT3MD_ENV_NOTE (1 << 31)

typedef struct {
	struct env_chan_dat{
		unsigned int period;
		unsigned int orn;
		unsigned int type;
		
		unsigned int noise;

		unsigned int cmd;
		unsigned int subparm;
		unsigned int parm;
	} env_chan;
	struct chan_dat {
		unsigned int note;
		
		unsigned int ins;
		unsigned int orn;
		unsigned int env;
		unsigned int vol;

		unsigned int cmd;
		unsigned int subparm;
		unsigned int parm;
	} chan[3];
} pt3md_pattern_data;

#include "pmd_events.h"

#pragma pack(pop)

// pt3.5+ tone tables
extern const u16 pt3_tonetab[4][12];

// very important(!) global pmd pos variable
extern volatile size_t pmd_pos;

// pt3 pointers
extern unsigned char *pt3file;
extern pt3_hdr *pt3hdr;

// pmd headers
extern pt3md_hdr *pmdhdr;
extern pt3md_mdl_hdr *pmdmdl;

// global data
struct pmdins_t {
	bool avail;
	pt3md_ins_hdr hdr;
	std::vector<u8> data; //      -      4,
						  //    ,    
};

struct pmdorn_t {
	bool avail;
	pt3md_orn hdr;
	std::vector<i8> data;
};

// hashed data stuff
struct hdat {
	std::vector<u8> p;
	unsigned long hash;
	size_t optidx;		// index in optimized table or data pos
};

struct pmdpat_t {
	bool avail;
	pt3md_pattern_pchan pchan;
	pt3md_pattern hdr;
	hdat data[4], autospd;
};

extern pmdins_t pmdins[32];
extern pmdorn_t pmdorn[16];

extern pmdpat_t pt3pat[MAX_PATTERNS];
extern std::vector<pmdpat_t> pmdpat;

extern u8 pt3ord[256];
extern std::vector<u8> pmdord;

// optimized data table
extern std::vector<hdat> optpat, optasd;
extern size_t optpos, asdpos;

extern bool verbose;

#include "optimize.h"