// pt3md v 0.3 header file - by wbcbz7 31.o5.2o17

#ifndef __PT3MD_H
#define __PT3MD_H

#ifndef u8
#define u8  unsigned char
#endif

#ifndef u16
#define u16 unsigned short
#endif

#ifndef i8
#define i8  signed char
#endif

#ifndef i16
#define i16 signed short
#endif

#define PT3MD_TYPE_3CH    0
#define PT3MD_TYPE_6CH    1
#define PT3MD_TYPE_9CH    2
#define PT3MD_TYPE_12CH   3
#define PT3MD_TYPE_AYX    4

enum pt3md_mdl_hdr_chip_stereo {
    abc = 0, acb, bac, mono
};

enum pt3md_mdl_hdr_chip_type {
    ay = 0, ym
};

#pragma pack(push, 1)

typedef struct {
    u8  sign[4];        // сигнатура модуля - "PMD\x1A"
    struct {
        u8 minor:4;
        u8 major:4;
    } ver;              // версия формата модуля: 0x03 - v0.3
    struct {
        u8 chipnum:3;   // количество используемых чипов минус один
        u8 mch_mode:1;  // режим прогрывания:
                        // 0 - чипы играют синхронно, в контейнере хранится
                        //     один модуль (синхронный, sync)
                        // 1 - для каждого чипа играет отдельный модуль
                        //     (асинхронный, async)
    } type;
    u16 reserved;    // резерв
} pt3md_hdr;

typedef struct {
    u8  pat_num;        // количество паттернов (0 - нет)
    u8  ord_num;        // длина списка проигрывания (0 - нет)
    u8  ins_num;        // количество инструментов (0 - нет)
    u8  orn_num;        // количество орнаментов (0 - нет)
    
    u8  loop_pos;       // позиция рестарта модуля
    u8  speed;          // начальная скорость модуля
    
    struct {            // дополнительные флаги проигрывания
        u8 tremor:1;    // 0 - команда 6xxx (тремор) сбрасывает ком. 1\2\3
                        // 1 - действие ком. 1\2\3 идет параллельно ком. 6
        u8 res:7;       // резерв
    } flags;
    
    struct {            // данные чипа
        u8 chiptype:1;  // тип чипа (0 - AY-3-8910, 1 - YM2149F)
        u8 stereo:2;    // раскладка каналов (левый-центр-правый):
                        // 0 - ABC, 1 - ACB, 2 - BAC, 3 - моно
        u8 res:4;       // резерв
        u8 ayx:1;       // расширенные фичи AYX-32 (1 - поддержаны)
    } chip;
    
    u16 int_rate;       // 15625 / (частота прерывания в Гц)
    u16 clock;          // тактовая частота чипа в КГц: 0x6D6 = 1,75 МГц
    
    // указатели
    u16 p_name;         // ASCIIZ-строка названия модуля (*p_name == 0 - нет)
    u16 p_author;       // ASCIIZ-строка автора модуля (*p_author == 0 - нет)
    
    u16 p_pattab;       // таблица паттернов
    u16 p_order;        // список проигрывания
    u16 p_instab;       // таблица указателей на инструменты
    u16 p_orntab;       // таблица указателей на орнаменты
    
    u16 p_notetab;      // таблица из 12 значений (little-endian) периодов AY
                        // для первой октавы (C-1...B-1)

    u16 p_misc;         // дополнительная информация (*p_misc == 0 - нет)
} pt3md_mdl_hdr;

// заголовок инструмента
typedef struct {
    u8 len;             // длина инструмента (0 = 256!)
    u8 loop_start;      // начало release-цикла (конец на строке len-1)
    u8 linesize:7;      // длина одной строки данных в байтах
    u8 sustain:1;       // 1 - есть sustain-цикл, 0 - нет
} pt3md_ins_hdr;

typedef
    // непосредственные данные инструмента (для AY)
    // linesize = 0 -> пустой инструмент!
    // linesize = 4 -> хранятся громкость и смещения тона\шума
    // linesize = 5 -> + смещение огибающей
    // linesize = 6 -> + смещение скважности (только в AYX-режиме!) !<
    // linesize = 8 -> + смещение панорамы канала !<
	struct {
      //u16 t_ofs:12;   // знаковое 12-бит смещение тона
        u8  t_ofsl;     // разделено на два u8, т.к. многие компиляторы C\C++
                        // размещают битовые поля разных типов с начала байта
        u8  t_ofsh:4;
        u8  t_mask:1;   // маска тона (0 - разрешен)
        u8  t_acc:1;    // 1 - накапливать смещение тона, 0 - не накапливать
        u8  n_acc:1;    // 1 - накапливать смещение шума, 0 - не накапливать
        u8  n_mask:1;   // маска шума
        
        u8  n_ofs:5;    // знаковое 5-бит смещение шума
        u8  v_slide:2;  // инкремент\декремент громкости,
                        // 0,1 - без изменений, 2 - -1, 3 - +1
        u8  p_mode:1;   // 0 - pan_l\pan_r - смещения, 1 - установка базы !<
                        // если linesize < 8 - резерв !<
        
        u8  volume:4;   // значение громкости
        u8  e_mask:1;   // маска огибающей (1(!) - разрешена)
        u8  e_acc:1;    // аналогично t_acc для смещения огибающей
        u8  d_acc:1;    // аналогично t_acc для смещения скважности !<
        u8  p_acc:1;    // аналогично t_acc для смещения панорамы !<
        
        i8  e_ofs;      // знаковое 8-бит смещение огибающей  (если есть)
        i8  d_ofs;      // знаковое 8-бит смещение скважности (если есть) !<
        
        i8  pan_l;      // знаковое смещение громкости левого  канала     !<
        i8  pan_r;      // знаковое смещение громкости правого канала     !<
} pt3md_ins_data;

typedef struct {
    u8 loop1_start;     // начало sustain-цикла (если есть)
    u8 loop1_end;       // конец  sustain-цикла (если есть)
} pt3md_ins_ftr;

// заголовок орнамента
typedef struct {
    u8  len;            // длина орнамента (0 = 256!)
    u8  loop_start;     // начало цикла (конец на строке len-1)
} pt3md_orn;

// заголовок таблицы паттернов
typedef union {
    struct {
        u16 p_chan_a;   // указатель на данные канала A
        u16 p_chan_b;   // указатель на данные канала B
        u16 p_chan_c;   // указатель на данные канала C
        u16 p_chan_d;   // указатель на данные канала огибающей\шума
    };
    u16 pchan[4];
} pt3md_pattern_pchan;

typedef struct {
    u8  patlen;         // длина паттерна (0 == 256(!))
    
    struct {            // данные автоскорости (см. ниже)
        u8    len;      // длина паттерна автоскорости (0 == 256(!))
        u16   ofs;      // указатель на паттерн автоскорости
    } autospd;
} pt3md_pattern;

#pragma pack(pop)

#endif
